#https://tools.ietf.org/html/rfc2131
#https://tools.ietf.org/html/rfc1533
import io
import enum
import asyncio
import ipaddress

class DHCPOpcode(enum.Enum):
	BOOTREQUEST = 1
	BOOTREPLY   = 2

class DHCPFlags(enum.IntFlag):
	B = 0x8000

#https://tools.ietf.org/html/rfc1700
class DHCPHardwareType(enum.Enum):
	STRING = 0 #not in RFC, but from packet dissection it seems like this type indicates a string
	ETHERNET_10MB                        =  1  
	EXPERIMENTAL_ETHERNET_3MB            =  2  
	AMATEUR_RADIO_AX25                    =  3  
	PROTEON_PRONET_TOKEN_RING              =  4  
	CHAOS                                  =  5  
	IEEE_802_NETWORKS                      =  6  
	ARCNET                                 =  7  
	HYPERCHANNEL                           =  8  
	LANSTAR                                =  9  
	AUTONET_SHORT_ADDRESS                  = 10  
	LOCALTALK                              = 11  
	LOCALNET                          = 12  
	ULTRA_LINK                             = 13  
	SMDS                                   = 14  
	FRAME_RELAY                            = 15  
	ASYNCHRONOUS_TRANSMISSION_MODE_16 = 16  
	HDLC                           = 17  
	FIBRE_CHANNEL                  = 18  
	ASYNCHRONOUS_TRANSMISSION_MODE_19 = 19  
	SERIAL_LINE                    = 20  
	ASYNCHRONOUS_TRANSMISSION_MODE_21 = 21 

def bytes_to_mac(b):
	return ':'.join([b.hex()[i:i+2] for i in range(0, len(b.hex()), 2)])

def mac_to_bytes(m):
	return bytes.fromhex(m.replace(':','').strip())

class DHCPMessage():
	def __init__(self):
		self.op = None
		self.htype = None #Hardware address type
		self.hlen  = None #Hardware address length
		self.hops  = None #Client sets to zero, optionally used by relay agents when booting via a relay agent.
		self.xid   = None #Transaction ID
		self.secs  = None #seconds elapsed since client began address acquisition or renewal process.
		self.flags = None #flags
		self.ciaddr = None #Client IP address
		self.yiaddr = None #'your' (client) IP address.
		self.siaddr = None #IP address of next server to use in bootstrap
		self.giaddr = None #Relay agent IP address
		self.chaddr = None #Client hardware address.
		self.chaddr_padding = None
		self.sname = None #Optional server host name, null terminated string.
		self.file = None #Boot file name, null terminated string;
		self.magic = None #Queen
		self.options = None
		self.padding = None

		#helper variables, not part of the standard
		self.dhcpmessagetype = None

	async def from_streamreader(reader):
		#running on UDP with no fragmentation possible, we just read everything from buffer and parse it
		data = await reader.read()
		return DHCPMessage.from_bytes(data)

	def from_bytes(bbuff):
		return DHCPMessage.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		msg = DHCPMessage()
		msg.op      = DHCPOpcode(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		msg.htype   = DHCPHardwareType(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		msg.hlen    = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		msg.hops    = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		msg.xid     = buff.read(4)
		msg.secs    = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		msg.flags   = DHCPFlags(int.from_bytes(buff.read(2), byteorder = 'big', signed=False))
		msg.ciaddr  = ipaddress.ip_address(buff.read(4))
		msg.yiaddr  = ipaddress.ip_address(buff.read(4))
		msg.siaddr  = ipaddress.ip_address(buff.read(4))
		msg.giaddr  = ipaddress.ip_address(buff.read(4))
		msg.chaddr  = bytes_to_mac(buff.read(msg.hlen))
		msg.chaddr_padding = buff.read(16-msg.hlen)
		msg.sname   = buff.read(64).decode().strip() #null-terminated
		msg.file    = buff.read(128).decode().strip() #null-terminated
		msg.magic   = buff.read(4)
		msg.options = DHCPOptionsParser.from_buffer(buff)
		msg.padding = buff.read()

		for option in msg.options:
			if option.code == 53:
				msg.dhcpmessagetype = option.msgtype

		return msg

	def construct(tid, opcode, options, hwtype = DHCPHardwareType.ETHERNET_10MB, 
					macaddress = 'AA:AA:AA:AA:AA:AA', hops = 0, secs = 0, 
					flags = DHCPFlags.B,
					giaddr = ipaddress.ip_address('0.0.0.0'), 
					ciaddr=ipaddress.ip_address('0.0.0.0'), 
					yiaddr=ipaddress.ip_address('0.0.0.0'), 
					siaddr=ipaddress.ip_address('0.0.0.0'),
					sname = '', file=''):
		msg = DHCPMessage()
		msg.op      = opcode
		msg.htype   = hwtype
		msg.chaddr  = macaddress
		msg.hlen    = len(mac_to_bytes(macaddress))
		msg.hops    = hops
		msg.xid     = tid
		msg.secs    = secs
		msg.flags   = flags
		msg.ciaddr  = ciaddr
		msg.yiaddr  = yiaddr
		msg.siaddr  = siaddr
		msg.giaddr  = giaddr
		msg.sname   = sname
		msg.file    = file
		msg.magic   = b'\x63\x82\x53\x63'
		msg.options = options
		return msg

	def to_bytes(self):
		t  = self.op.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.htype.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.hlen.to_bytes(1, byteorder = 'big', signed = False)
		t += self.hops.to_bytes(1, byteorder = 'big', signed = False)
		t += self.xid
		t += self.secs.to_bytes(2, byteorder = 'big', signed = False)
		t += self.flags.to_bytes(2, byteorder = 'big', signed = False)
		t += self.ciaddr.packed
		t += self.yiaddr.packed 
		t += self.siaddr.packed 
		t += self.giaddr.packed 
		t += mac_to_bytes(self.chaddr) + b'\x00'*(16-self.hlen)
		t += self.sname.encode() + b'\x00'*(64-len(self.sname.encode()))
		t += self.file.encode() + b'\x00'*(128-len(self.file.encode()))
		t += self.magic
		for option in self.options:
			t += option.to_bytes()
		
		t += b'\x00'*(576 - len(t))

		return t

	def __repr__(self):
		t  = '== DHCPMessage ==\r\n'
		t += 'op : %s\r\n' % self.op
		t += 'htype : %s\r\n' % self.htype
		t += 'hlen : %s\r\n' % self.hlen
		t += 'hops : %s\r\n' % self.hops
		t += 'xid : %s\r\n' % repr(self.xid)
		t += 'secs : %s\r\n' % self.secs
		t += 'flags : %s\r\n' % repr(self.flags.name)
		t += 'ciaddr : %s\r\n' % self.ciaddr
		t += 'yiaddr : %s\r\n' % self.yiaddr
		t += 'siaddr : %s\r\n' % self.siaddr
		t += 'giaddr : %s\r\n' % self.giaddr
		t += 'chaddr : %s\r\n' % self.chaddr
		t += 'chaddr_padding : %s\r\n' % self.chaddr_padding
		t += 'sname : %s\r\n' % self.sname
		t += 'file : %s\r\n' % self.file
		t += 'magic : %s\r\n' % self.magic
		for option in self.options:
			t += repr(option)

		return t

class DHCPOptPAD():
	def __init__(self):
		self.code = 0

	def from_buffer(buff):
		opt = DHCPOptPAD()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		return opt
	def construct():
		opt = DHCPOptPAD()
		return opt
	def to_bytes(self):
		return self.code.to_bytes(1, byteorder = 'big', signed = False)

	def __repr__(self):
		t = '= PAD =\r\n'
		return t

class DHCPOptEND():
	def __init__(self):
		self.code = 255

	def from_buffer(buff):
		opt = DHCPOptEND()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		return opt

	def construct():
		opt = DHCPOptEND()
		return opt

	def to_bytes(self):
		return self.code.to_bytes(1, byteorder = 'big', signed = False)

	def __repr__(self):
		t = '= END =\r\n'
		return t

class DHCPOptSUBNETMASK():
	def __init__(self):
		self.code = 1
		self.len  = 4
		self.mask = None

	def from_buffer(buff):
		opt = DHCPOptSUBNETMASK()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.mask = bytes_to_mac(buff.read(opt.len))
		return opt

	def construct(mask):
		opt = DHCPOptSUBNETMASK()
		opt.mask = mask
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += mac_to_bytes(self.mask)
		return t

	def __repr__(self):
		t  = '= SUBNETMASK =\r\n'
		t += 'NetMask: %s\r\n' % self.mask
		return t

class DHCPOptTIMEOFFSET():
	def __init__(self):
		self.code = 2
		self.len  = 4
		self.timeoffset = None
	
	def from_buffer(buff):
		opt = DHCPOptTIMEOFFSET()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.timeoffset = int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=True)
		return opt

	def construct(timeoffset):
		opt = DHCPOptTIMEOFFSET()
		opt.timeoffset = timeoffset
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.timeoffset.to_bytes(4,byteorder = 'big', signed=True)
		return t

	def __repr__(self):
		t  = '= TIMEOFFSET =\r\n'
		t += 'timeoffset: %s\r\n' % self.timeoffset
		return t

class DHCPOptROUTERS():
	def __init__(self):
		self.code = 3
		self.len  = None
		self.addresses = []
	
	def from_buffer(buff):
		opt = DHCPOptROUTERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptROUTERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= ROUTERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptTIMESERVERS():
	def __init__(self):
		self.code = 4
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptTIMESERVERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptTIMESERVERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= TIMESERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptNAMESERVERS():
	def __init__(self):
		self.code = 5
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptNAMESERVERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptNAMESERVERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= NAMESERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptDNSSERVERS():
	def __init__(self):
		self.code = 6
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptDNSSERVERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptDNSSERVERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def from_setting(setting):
		opt = DHCPOptDNSSERVERS()
		addresses = setting
		if not isinstance(setting, list):
			addresses = [setting]
		for address in addresses:
			opt.addresses.append(ipaddress.ip_address(address))
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= DNSSERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t


class DHCPOptLOGSERVERS():
	def __init__(self):
		self.code = 7
		self.len  = None
		self.addresses = []
	
	def from_buffer(buff):
		opt = DHCPOptLOGSERVERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptLOGSERVERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= LOGSERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptCOOKIESERVER():
	def __init__(self):
		self.code = 8
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptCOOKIESERVER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptCOOKIESERVER()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= COOKIESERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptLPRSERVERS():
	def __init__(self):
		self.code = 9
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptLPRSERVERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptLPRSERVERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= LPRSERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptIMPRESSSERVERS():
	def __init__(self):
		self.code = 10
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptIMPRESSSERVERS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptIMPRESSSERVERS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= IMPRESSSERVERS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptRESOURCELOCATIONS():
	def __init__(self):
		self.code = 11
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptRESOURCELOCATIONS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptRESOURCELOCATIONS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= RESOURCELOCATIONS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptHOSTNAME():
	def __init__(self):
		self.code = 12
		self.len  = None
		self.hostname = None

	def from_buffer(buff):
		opt = DHCPOptHOSTNAME()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.hostname = buff.read(opt.len).decode()
		return opt

	def construct(hostname):
		opt = DHCPOptHOSTNAME()
		opt.hostname = hostname
		opt.len = len(hostname.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.hostname.encode()
		return t

	def __repr__(self):
		t  = '= HOSTNAME =\r\n'
		t += 'hostname: %s\r\n' % str(self.hostname)
		return t

class DHCPOptBOOTFILESIZE():
	def __init__(self):
		self.code = 13
		self.len  = 2
		self.filesize = None

	def from_buffer(buff):
		opt = DHCPOptBOOTFILESIZE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.filesize = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		return opt

	def construct(filesize):
		opt = DHCPOptBOOTFILESIZE()
		opt.filesize = filesize
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.filesize.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= BOOTFILESIZE =\r\n'
		t += 'filesize: %s\r\n' % str(self.filesize)
		return t


class DHCPOptMERITDUMPFILE():
	def __init__(self):
		self.code = 14
		self.len  = None
		self.pathname = None

	def from_buffer(buff):
		opt = DHCPOptMERITDUMPFILE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.pathname = buff.read(opt.len).decode()
		return opt

	def construct(pathname):
		opt = DHCPOptMERITDUMPFILE()
		opt.pathname = pathname
		opt.len = len(pathname.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.pathname.encode()
		return t

	def __repr__(self):
		t  = '= MERITDUMPFILE =\r\n'
		t += 'pathname: %s\r\n' % str(self.pathname)
		return t


class DHCPOptDOMAINNAME():
	def __init__(self):
		self.code = 15
		self.len  = None
		self.domainname = None

	def from_buffer(buff):
		opt = DHCPOptDOMAINNAME()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.domainname = buff.read(opt.len).decode()
		return opt

	def construct(domainname):
		opt = DHCPOptDOMAINNAME()
		opt.domainname = domainname
		opt.len = len(domainname.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.domainname.encode()
		return t

	def __repr__(self):
		t  = '= DOMAINNAME =\r\n'
		t += 'domainname: %s\r\n' % str(self.domainname)
		return t

class DHCPOptSWAPSERVER():
	def __init__(self):
		self.code = 16
		self.len  = 4
		self.address = None

	def from_buffer(buff):
		opt = DHCPOptSWAPSERVER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.address = ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
		return opt

	def construct(address):
		opt = DHCPOptSWAPSERVER()
		opt.address = address
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.address.packed
		return t

	def __repr__(self):
		t  = '= SWAPSERVER =\r\n'
		t += 'address: %s\r\n' % str(self.address)
		return t

class DHCPOptROOTPATH():
	def __init__(self):
		self.code = 17
		self.len  = None
		self.pathname = None

	def from_buffer(buff):
		opt = DHCPOptROOTPATH()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.pathname = buff.read(opt.len).decode()
		return opt

	def construct(pathname):
		opt = DHCPOptROOTPATH()
		opt.pathname = pathname
		opt.len = len(pathname.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.pathname.encode()
		return t

	def __repr__(self):
		t  = '= ROOTPATH =\r\n'
		t += 'pathname: %s\r\n' % str(self.pathname)
		return t

class DHCPOptEXTENSIONSPATH():
	def __init__(self):
		self.code = 18
		self.len  = None
		self.pathname = None

	def from_buffer(buff):
		opt = DHCPOptEXTENSIONSPATH()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.pathname = buff.read(opt.len).decode()
		return opt

	def construct(pathname):
		opt = DHCPOptEXTENSIONSPATH()
		opt.pathname = pathname
		opt.len = len(pathname.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.pathname.encode()
		return t

	def __repr__(self):
		t  = '= EXTENSIONSPATH =\r\n'
		t += 'pathname: %s\r\n' % str(self.pathname)
		return t

class DHCPOptIPFORWARDING():
	def __init__(self):
		self.code = 19
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptIPFORWARDING()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptIPFORWARDING()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= IPFORWARDING =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptNONLOCALSRCROUTING():
	def __init__(self):
		self.code = 20
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptNONLOCALSRCROUTING()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptNONLOCALSRCROUTING()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= NONLOCALSRCROUTING =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptPOLICYFILTER():
	def __init__(self):
		self.code = 21
		self.len  = None
		self.destmasks = [] #should be a list of tuples

	def from_buffer(buff):
		opt = DHCPOptPOLICYFILTER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/8)):
			ip = ipaddress.ip_addres(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
			mask = bytes_to_mac(buff.read(4))
			opt.destmasks.append((ip,mask))
		
		return opt

	def construct(destmasks):
		opt = DHCPOptPOLICYFILTER()
		if not isinsance(destmasks,list):
			destmasks = [destmasks]
		opt.destmasks = destmasks
		opt.len = len(destmasks)*8

		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for destmask in destmasks:
			t += destmask[0].packed + mac_to_bytes(destmask[1])
		return t

	def __repr__(self):
		t  = '= POLICYFILTER =\r\n'
		for addr, mask in self.destmasks:
			t += 'addr: %s mask:%s\r\n' % (str(addr), mask)
		return t

class DHCPOptMAXIMUMDATAGRAMREASSEMBLY():
	def __init__(self):
		self.code = 22
		self.len  = 2
		self.size = None

	def from_buffer(buff):
		opt = DHCPOptMAXIMUMDATAGRAMREASSEMBLY()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.size = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		return opt

	def construct(size):
		opt = DHCPOptMAXIMUMDATAGRAMREASSEMBLY()
		opt.size = size
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.size.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= MAXIMUMDATAGRAMREASSEMBLY =\r\n'
		t += 'size: %s\r\n' % str(self.size)
		return t

class DHCPOptDEFAULTTTL():
	def __init__(self):
		self.code = 23
		self.len  = 1
		self.ttl  = None

	def from_buffer(buff):
		opt = DHCPOptDEFAULTTTL()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.ttl = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		return opt

	def construct(ttl):
		opt = DHCPOptDEFAULTTTL()
		opt.ttl = ttl
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.ttl.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= DEFAULTTTL =\r\n'
		t += 'ttl: %s\r\n' % str(self.ttl)
		return t

class DHCPOptPATHMTUAGINGTIMEOUT():
	def __init__(self):
		self.code = 24
		self.len  = 4
		self.timeout  = None

	def from_buffer(buff):
		opt = DHCPOptPATHMTUAGINGTIMEOUT()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.timeout = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(ttl):
		opt = DHCPOptPATHMTUAGINGTIMEOUT()
		opt.timeout = timeout
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.timeout.to_bytes(4, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= PATHMTUAGINGTIMEOUT =\r\n'
		t += 'timeout: %s\r\n' % str(self.timeout)
		return t

class DHCPOptMTUPLATEUTABLE():
	def __init__(self):
		self.code = 25
		self.len  = None
		self.sizes= [] #list of sizes (of 2byte integers)

	def from_buffer(buff):
		opt = DHCPOptMTUPLATEUTABLE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(opt.len/2):
			opt.sizes.append(int.from_bytes(buff.read(2), byteorder = 'big', signed=False))
		return opt

	def construct(sizes):
		opt = DHCPOptMTUPLATEUTABLE()
		opt.sizes = sizes
		opt.len = len(sizes) * 2
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for size in sizes:
			t += size.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= MTUPLATEUTABLE =\r\n'
		t += 'sizes: %s\r\n' % str(self.sizes)
		return t

class DHCPOptINTERFACEMTU():
	def __init__(self):
		self.code = 26
		self.len  = 2
		self.mtu  = None

	def from_buffer(buff):
		opt = DHCPOptINTERFACEMTU()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.mtu = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		return opt

	def construct(mtu):
		opt = DHCPOptINTERFACEMTU()
		opt.mtu = mtu
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.mtu.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= INTERFACEMTU =\r\n'
		t += 'mtu: %s\r\n' % str(self.mtu)
		return t

class DHCPOptALLSUBNETSARELOCAL():
	def __init__(self):
		self.code = 27
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptALLSUBNETSARELOCAL()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptALLSUBNETSARELOCAL()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= ALLSUBNETSARELOCAL =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t


class DHCPOptBROADCASTADDRESS():
	def __init__(self):
		self.code = 28
		self.len  = 4
		self.address = None

	def from_buffer(buff):
		opt = DHCPOptBROADCASTADDRESS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.address = ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
		return opt

	def construct(address):
		opt = DHCPOptBROADCASTADDRESS()
		opt.address = address
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.address.packed
		return t

	def __repr__(self):
		t  = '= BROADCASTADDRESS =\r\n'
		t += 'address: %s\r\n' % str(self.address)
		return t

class DHCPOptPERFORMMASKDISCOVERY():
	def __init__(self):
		self.code = 29
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptPERFORMMASKDISCOVERY()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptPERFORMMASKDISCOVERY()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= PERFORMMASKDISCOVERY =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptMASKSUPPLIER():
	def __init__(self):
		self.code = 30
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptMASKSUPPLIER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptMASKSUPPLIER()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= MASKSUPPLIER =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t


class DHCPOptPERFORMROUTERDISCOVERY():
	def __init__(self):
		self.code = 31
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptPERFORMROUTERDISCOVERY()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptPERFORMROUTERDISCOVERY()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t
	
	def __repr__(self):
		t  = '= PERFORMROUTERDISCOVERY =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptROUTERSOLICITATIONADDRESS():
	def __init__(self):
		self.code = 32
		self.len  = 4
		self.address = None

	def from_buffer(buff):
		opt = DHCPOptROUTERSOLICITATIONADDRESS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.address = ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
		return opt

	def construct(address):
		opt = DHCPOptROUTERSOLICITATIONADDRESS()
		opt.address = address
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.address.packed
		return t

	def __repr__(self):
		t  = '= ROUTERSOLICITATIONADDRESS =\r\n'
		t += 'address: %s\r\n' % str(self.address)
		return t

class DHCPOptSTATICROUTES():
	def __init__(self):
		self.code = 33
		self.len  = None
		self.addressroutes = []

	def from_buffer(buff):
		opt = DHCPOptSTATICROUTES()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/8)):
			ip = ipaddress.ip_addres(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
			route = ipaddress.ip_addres(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
			opt.addressroutes.append((ip,route))
		
		return opt

	def construct(addressroutes):
		opt = DHCPOptSTATICROUTES()
		if not isinsance(addressroutes,list):
			addressroutes = [addressroutes]
		opt.addressroutes = addressroutes
		opt.len = len(addressroutes) * 8

		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for addressroutes in self.addressroutes:
			t += addressroute[0].packed + addressroute[1].packed
		return t

	def __repr__(self):
		t  = '= STATICROUTES =\r\n'
		for addr, route in self.addressroutes:
			t += 'addr: %s route: %s \r\n' % (str(addr),str(route))
		return t

class DHCPOptTRAILERENCAPSULATION():
	def __init__(self):
		self.code = 34
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptTRAILERENCAPSULATION()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptTRAILERENCAPSULATION()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= TRAILERENCAPSULATION =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptARPCACHETIMEOUT():
	def __init__(self):
		self.code = 35
		self.len  = 4
		self.timeout = None

	def from_buffer(buff):
		opt = DHCPOptARPCACHETIMEOUT()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.ttl = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(ttl):
		opt = DHCPOptARPCACHETIMEOUT()
		opt.timeout = timeout
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.timeout.to_bytes(4, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= ARPCACHETIMEOUT =\r\n'
		t += 'timeout: %s\r\n' % str(self.timeout)
		return t

class DHCPOptETHERNETENCAPSULATION():
	def __init__(self):
		self.code = 36
		self.len  = 1
		self.enabled = None

	def from_buffer(buff):
		opt = DHCPOptETHERNETENCAPSULATION()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptETHERNETENCAPSULATION()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= ETHERNETENCAPSULATION =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptTCPDEFAULTTTL():
	def __init__(self):
		self.code = 37
		self.len  = 1
		self.ttl = None
	
	def from_buffer(buff):
		opt = DHCPOptTCPDEFAULTTTL()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.ttl = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(ttl):
		opt = DHCPOptTCPDEFAULTTTL()
		opt.timeout = timeout
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.timeout.to_bytes(4, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= TCPDEFAULTTTL =\r\n'
		t += 'ttl: %s\r\n' % str(self.ttl)
		return t

class DHCPOptTCPKEEPALIVEINTERVAL():
	def __init__(self):
		self.code = 38
		self.len  = 4
		self.timeout = None 

	def from_buffer(buff):
		opt = DHCPOptTCPKEEPALIVEINTERVAL()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.ttl = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(ttl):
		opt = DHCPOptTCPKEEPALIVEINTERVAL()
		opt.timeout = timeout
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.timeout.to_bytes(4, byteorder = 'big', signed = False)
		return t
	
	def __repr__(self):
		t  = '= TCPKEEPALIVEINTERVAL =\r\n'
		t += 'timeout: %s\r\n' % str(self.timeout)
		return t

class DHCPOptTCPKEEPALIVEGARBAGE():
	def __init__(self):
		self.code = 39
		self.len  = 1
		self.enabled  = None

	def from_buffer(buff):
		opt = DHCPOptTCPKEEPALIVEGARBAGE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.enabled = bool(int.from_bytes(buff.read(opt.len), byteorder = 'big', signed=False))
		return opt

	def construct(enabled):
		opt = DHCPOptTCPKEEPALIVEGARBAGE()
		opt.enabled = enabled
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.enabled.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= TCPKEEPALIVEGARBAGE =\r\n'
		t += 'enabled: %s\r\n' % str(self.enabled)
		return t

class DHCPOptNISDOMAIN():
	def __init__(self):
		self.code = 40
		self.len  = None
		self.domainname = None

	def from_buffer(buff):
		opt = DHCPOptNISDOMAIN()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.domainname = buff.read(opt.len).decode()
		return opt

	def construct(domainname):
		opt = DHCPOptNISDOMAIN()
		opt.domainname = domainname
		opt.len = len(domainname.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.domainname.encode()
		return t
	
	def __repr__(self):
		t  = '= NISDOMAIN =\r\n'
		t += 'domainname: %s\r\n' % str(self.domainname)
		return t

class DHCPOptNIS():
	def __init__(self):
		self.code = 41
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptNIS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptNIS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= NIS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptNTP():
	def __init__(self):
		self.code = 42
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptNTP()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def from_setting(setting):
		opt = DHCPOptNTP()
		addresses = setting
		if not isinstance(setting, list):
			addresses = [setting]
		for address in addresses:
			opt.addresses.append(ipaddress.ip_address(address))
		opt.len = len(addresses) *4
		return opt

	def construct(addresses):
		opt = DHCPOptNTP()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= NTP =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptVENDORSPECIFIC():
	def __init__(self):
		self.code = 43
		self.len  = None
		self.data = None

	def from_buffer(buff):
		opt = DHCPOptVENDORSPECIFIC()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.data = buff.read(opt.len)
		return opt

	def __repr__(self):
		t  = '= VENDORSPECIFIC =\r\n'
		t += 'code: %s\r\n' % str(self.code)
		t += 'len: %s\r\n' % str(self.len)
		t += 'data: %s\r\n' % str(self.data.hex())
		return t

class DHCPOptNETBIOSOVERTCP():
	def __init__(self):
		self.code = 44
		self.len  = None
		self.addresses = []
	
	def from_buffer(buff):
		opt = DHCPOptNETBIOSOVERTCP()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptNETBIOSOVERTCP()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t
	def __repr__(self):
		t  = '= NETBIOSOVERTCP =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptNETBIOSOVERTCPDDS():
	def __init__(self):
		self.code = 45
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptNETBIOSOVERTCPDDS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptNETBIOSOVERTCPDDS()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= NETBIOSOVERTCPDDS =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DNSOptNetBIOSNodeType(enum.Enum):
	B_NODE = 1
	P_NODE = 2
	M_NODE = 4
	H_NODE = 8

class DHCPOptNETBIOSOVERTCPNODETYPE():
	def __init__(self):
		self.code = 46
		self.len  = 1
		self.nodetype = None
	
	def from_buffer(buff):
		opt = DHCPOptNETBIOSOVERTCPNODETYPE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.nodetype = DNSOptNetBIOSNodeType(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		return opt

	def construct(nodetype):
		opt = DHCPOptNETBIOSOVERTCPNODETYPE()
		if not isinstance(nodetype, DNSOptNetBIOSNodeType):
			nodetype = DNSOptNetBIOSNodeType[nodetype]
		opt.nodetype = nodetype
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.nodetype.value.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= NETBIOSOVERTCPNODETYPE =\r\n'
		t += 'nodetype: %s\r\n' % str(self.nodetype)
		return t

class DHCPOptNETBIOSOVERTCPSCOPE():
	#TODO
	def __init__(self):
		self.code = 47
		self.len  = None
		self.scope = None

	def from_buffer(buff):
		opt = DHCPOptNETBIOSOVERTCPSCOPE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.scope = buff.read(opt.len)
		return opt

	def construct(scope):
		opt = DHCPOptNETBIOSOVERTCPSCOPE()
		opt.scope = scope
		opt.len = len(scope)
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.scope
		return t

	def __repr__(self):
		t  = '= NETBIOSOVERTCPSCOPE =\r\n'
		t += 'scope: %s\r\n' % str(self.scope)
		return t

class DHCPOptXWINDOWFONTSERVER():
	def __init__(self):
		self.code = 48
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptXWINDOWFONTSERVER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptXWINDOWFONTSERVER()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= XWINDOWFONTSERVER =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptXWINDOWDISPLAYMANAGER():
	def __init__(self):
		self.code = 49
		self.len  = None
		self.addresses = []

	def from_buffer(buff):
		opt = DHCPOptXWINDOWDISPLAYMANAGER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(int(opt.len/4)):
			opt.addresses.append(ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False)))
		return opt

	def construct(addresses):
		opt = DHCPOptXWINDOWDISPLAYMANAGER()
		if not isinstance(addresses, list):
			addresses = [addresses]
		opt.addresses = addresses
		opt.len = len(addresses) *4
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for address in self.addresses:
			t += address.packed
		return t

	def __repr__(self):
		t  = '= XWINDOWDISPLAYMANAGER =\r\n'
		for addr in self.addresses:
			t += 'addr: %s\r\n' % str(addr)
		return t

class DHCPOptREQUESTEDIPADDRESS():
	def __init__(self):
		self.code = 50
		self.len  = 4
		self.address = None

	def from_buffer(buff):
		opt = DHCPOptREQUESTEDIPADDRESS()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.address = ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
		return opt

	def construct(address):
		opt = DHCPOptREQUESTEDIPADDRESS()
		opt.address = address
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.address.packed
		return t

	def __repr__(self):
		t  = '= REQUESTEDIPADDRESS =\r\n'
		t += 'address: %s\r\n' % str(self.address)
		return t

class DHCPOptIPADDRESSLEASETIME():
	def __init__(self):
		self.code = 51
		self.len  = 4
		self.leasetime = None


	def from_buffer(buff):
		opt = DHCPOptIPADDRESSLEASETIME()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.leasetime = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(leasetime):
		opt = DHCPOptIPADDRESSLEASETIME()
		opt.leasetime = leasetime
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.leasetime.to_bytes(4, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= IPADDRESSLEASETIME =\r\n'
		t += 'leasetime: %s\r\n' % str(self.leasetime)
		return t

class DNSOptOptionOverLoadType(enum.Enum):
	FILE = 1
	SNAME = 2
	BOTH = 3

class DHCPOptOPTIONOVERLOAD():
	def __init__(self):
		self.code = 52
		self.len  = 1
		self.value = None

	def from_buffer(buff):
		opt = DHCPOptOPTIONOVERLOAD()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.value = DNSOptOptionOverLoadType(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		return opt

	def construct(value):
		opt = DHCPOptOPTIONOVERLOAD()
		opt.value = value
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.value.value.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= OPTIONOVERLOAD =\r\n'
		t += 'value: %s\r\n' % str(self.value)
		return t


class DHCPOptMessageType(enum.Enum):
	DHCPDISCOVER = 1
	DHCPOFFER    = 2
	DHCPREQUEST  = 3
	DHCPDECLINE  = 4
	DHCPACK      = 5
	DHCPNAK      = 6
	DHCPRELEASE  = 7
	DHCPINFORM   = 8


class DHCPOptDHCPMESSAGETYPE():
	def __init__(self):
		self.code = 53
		self.len  = 1
		self.msgtype = None

	def from_buffer(buff):
		opt = DHCPOptDHCPMESSAGETYPE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.msgtype = DHCPOptMessageType(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		return opt

	def construct(msgtype):
		opt = DHCPOptDHCPMESSAGETYPE()
		opt.msgtype = msgtype
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.msgtype.value.to_bytes(1, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= DHCPMESSAGETYPE =\r\n'
		t += 'msgtype: %s\r\n' % str(self.msgtype)
		return t

class DHCPOptSERVERIDENTIFIER():
	def __init__(self):
		self.code = 54
		self.len  = 4
		self.address = None
	
	def from_buffer(buff):
		opt = DHCPOptSERVERIDENTIFIER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.address = ipaddress.ip_address(int.from_bytes(buff.read(4), byteorder = 'big', signed=False))
		return opt

	def construct(address):
		opt = DHCPOptSERVERIDENTIFIER()
		opt.address = address
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.address.packed
		return t

	def __repr__(self):
		t  = '= DHCPMESSAGETYPE =\r\n'
		t += 'address: %s\r\n' % str(self.address)
		return t


class DHCPOptPARAMETERREQUEST():
	def __init__(self):
		self.code = 55
		self.len  = None
		self.optioncodes = []

	def from_buffer(buff):
		opt = DHCPOptPARAMETERREQUEST()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		for i in range(opt.len):
			opt.optioncodes.append(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		return opt

	def construct(optioncodes):
		opt = DHCPOptPARAMETERREQUEST()
		opt.optioncodes = optioncodes
		opt.len = len(optioncodes)
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		for optioncode in self.optioncodes:
			t+= optioncode.to_bytes(1, byteorder = 'big', signed = False)
		return t
	
	def __repr__(self):
		t  = '= PARAMETERREQUEST =\r\n'
		t += 'optioncodes: %s\r\n' % str(self.optioncodes)
		return t

class DHCPOptMESSAGE():
	def __init__(self):
		self.code = 56
		self.len  = None
		self.message = None

	def from_buffer(buff):
		opt = DHCPOptMESSAGE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.message = buff.read(opt.len).decode()
		return opt

	def construct(message):
		opt = DHCPOptMESSAGE()
		opt.message = message
		opt.len = len(message.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.message.encode()
		return t

	def __repr__(self):
		t  = '= MESSAGE =\r\n'
		t += 'message: %s\r\n' % str(self.message)
		return t

class DHCPOptMAXIMUMDHCPMESSAGESIZE():
	def __init__(self):
		self.code = 57
		self.len  = 2
		self.length = None

	def from_buffer(buff):
		opt = DHCPOptMAXIMUMDHCPMESSAGESIZE()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.length = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		return opt

	def construct(length):
		opt = DHCPOptMAXIMUMDHCPMESSAGESIZE()
		opt.length = length
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.length.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= MAXIMUMDHCPMESSAGESIZE =\r\n'
		t += 'length: %s\r\n' % str(self.length)
		return t

class DHCPOptRENEVALTIME():
	def __init__(self):
		self.code = 58
		self.len  = 4
		self.interval = None

	def from_buffer(buff):
		opt = DHCPOptRENEVALTIME()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.interval = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(interval):
		opt = DHCPOptRENEVALTIME()
		opt.interval = interval
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.interval.to_bytes(4, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= RENEVALTIME =\r\n'
		t += 'interval: %s\r\n' % str(self.interval)
		return t


class DHCPOptREBINDINGTIME():
	def __init__(self):
		self.code = 59
		self.len  = 4
		self.interval = None

	def from_buffer(buff):
		opt = DHCPOptREBINDINGTIME()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.interval = int.from_bytes(buff.read(4), byteorder = 'big', signed=False)
		return opt

	def construct(interval):
		opt = DHCPOptREBINDINGTIME()
		opt.interval = interval
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.interval.to_bytes(4, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '= REBINDINGTIME =\r\n'
		t += 'interval: %s\r\n' % str(self.interval)
		return t

class DHCPOptCLASSIDENTIFIER():
	def __init__(self):
		self.code = 60
		self.len  = None
		self.classid = None

	def from_buffer(buff):
		opt = DHCPOptCLASSIDENTIFIER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.classid = buff.read(opt.len).decode()
		return opt

	def construct(classid):
		opt = DHCPOptCLASSIDENTIFIER()
		opt.classid = classid
		opt.len = len(opt.classid.encode())
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.classid.encode()
		return t

	def __repr__(self):
		t  = '= CLASSIDENTIFIER =\r\n'
		t += 'classid: %s\r\n' % str(self.classid)
		return t

class DHCPOptCLIENTIDENTIFIER():
	def __init__(self):
		self.code = 61
		self.len  = None
		self.type = None
		self.clientid = None

	def from_buffer(buff):
		opt = DHCPOptCLIENTIDENTIFIER()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.type = DHCPHardwareType(int.from_bytes(buff.read(1), byteorder = 'big', signed=False))
		opt.clientid = buff.read(opt.len-1)
		return opt

	def construct(clienttype, clientid):
		opt = DHCPOptCLIENTIDENTIFIER()
		opt.clientid = clientid
		opt.len = len(opt.clientid)-1
		opt.type = clienttype
		return opt

	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.type.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.clientid
		return t

	def __repr__(self):
		t  = '= CLIENTIDENTIFIER =\r\n'
		t += 'clientid: %s\r\n' % str(self.clientid)
		return t

class DHCPOptXXX():
	def __init__(self):
		self.code = None
		self.len  = None
		self.data = None

	def from_buffer(buff):
		opt = DHCPOptXXX()
		opt.code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.len = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
		opt.data = buff.read(opt.len)
		return opt
		
	def to_bytes(self):
		t  = self.code.to_bytes(1, byteorder = 'big', signed = False)
		t += self.len.to_bytes(1, byteorder = 'big', signed = False)
		t += self.data
		return t

	def __repr__(self):
		t  = '= XXX =\r\n'
		t += 'code: %s\r\n' % str(self.code)
		t += 'len: %s\r\n' % str(self.len)
		t += 'data: %s\r\n' % str(self.data.hex())
		return t

OPTCode2ClassName = {
	0  : DHCPOptPAD,
	1  : DHCPOptSUBNETMASK,
	2  : DHCPOptTIMEOFFSET,
	3  : DHCPOptROUTERS,
	4  : DHCPOptTIMESERVERS,
	5  : DHCPOptNAMESERVERS,
	6  : DHCPOptDNSSERVERS,
	7  : DHCPOptLOGSERVERS,
	8  : DHCPOptCOOKIESERVER,
	9  : DHCPOptLPRSERVERS,
	10 : DHCPOptIMPRESSSERVERS,
	11 : DHCPOptRESOURCELOCATIONS,
	12 : DHCPOptHOSTNAME,
	13 : DHCPOptBOOTFILESIZE,
	14 : DHCPOptMERITDUMPFILE,
	15 : DHCPOptDOMAINNAME,
	16 : DHCPOptSWAPSERVER,
	17 : DHCPOptROOTPATH,
	18 : DHCPOptEXTENSIONSPATH,
	19 : DHCPOptIPFORWARDING,
	20 : DHCPOptNONLOCALSRCROUTING,
	21 : DHCPOptPOLICYFILTER,
	22 : DHCPOptMAXIMUMDATAGRAMREASSEMBLY,
	23 : DHCPOptDEFAULTTTL,
	24 : DHCPOptPATHMTUAGINGTIMEOUT,
	25 : DHCPOptMTUPLATEUTABLE,
	26 : DHCPOptINTERFACEMTU,
	27 : DHCPOptALLSUBNETSARELOCAL,
	28 : DHCPOptBROADCASTADDRESS,
	29 : DHCPOptPERFORMMASKDISCOVERY,
	30 : DHCPOptMASKSUPPLIER,
	31 : DHCPOptPERFORMROUTERDISCOVERY,
	32 : DHCPOptROUTERSOLICITATIONADDRESS,
	33 : DHCPOptSTATICROUTES,
	34 : DHCPOptTRAILERENCAPSULATION,
	35 : DHCPOptARPCACHETIMEOUT,
	36 : DHCPOptETHERNETENCAPSULATION,
	37 : DHCPOptTCPDEFAULTTTL,
	38 : DHCPOptTCPKEEPALIVEINTERVAL,
	39 : DHCPOptTCPKEEPALIVEGARBAGE,
	40 : DHCPOptNISDOMAIN,
	41 : DHCPOptNIS,
	42 : DHCPOptNTP,
	43 : DHCPOptVENDORSPECIFIC,
	44 : DHCPOptNETBIOSOVERTCP,
	45 : DHCPOptNETBIOSOVERTCPDDS,
	46 : DHCPOptNETBIOSOVERTCPNODETYPE,
	47 : DHCPOptNETBIOSOVERTCPSCOPE,
	48 : DHCPOptXWINDOWFONTSERVER,
	49 : DHCPOptXWINDOWDISPLAYMANAGER,
	50 : DHCPOptREQUESTEDIPADDRESS,
	51 : DHCPOptIPADDRESSLEASETIME,
	52 : DHCPOptOPTIONOVERLOAD,
	53 : DHCPOptDHCPMESSAGETYPE,
	54 : DHCPOptSERVERIDENTIFIER,
	55 : DHCPOptPARAMETERREQUEST,
	56 : DHCPOptMESSAGE,
	57 : DHCPOptMAXIMUMDHCPMESSAGESIZE,
	58 : DHCPOptRENEVALTIME,
	59 : DHCPOptREBINDINGTIME,
	60 : DHCPOptCLASSIDENTIFIER,
	61 : DHCPOptCLIENTIDENTIFIER,
	255: DHCPOptEND
}

class DHCPOptionsParser():
	def __init__(self):
		pass

	def from_buffer(buff):
		options = []
		while True:
			pos = buff.tell()
			code = int.from_bytes(buff.read(1), byteorder = 'big', signed=False)
			buff.seek(pos, io.SEEK_SET)
			if code in OPTCode2ClassName:
				opt = OPTCode2ClassName[code].from_buffer(buff)
				options.append(opt)
				if opt.code == 255:
					break
			else:
				opt = DHCPOptXXX.from_buffer(buff)
				options.append(opt)

		return options
		



import json
from responder3.core.commons import *

class NetworkInterface:

	def __init__(self):
		"""
		Container object to describe a network interface
		"""
		self.ifname = None
		self.ifindex = None # zone_indices in windows
		self.addresses = []
	
	def to_dict(self):
		return {
            "ifname": str(self.ifname),
            "ifindex": str(self.ifindex),
            "addresses": self.addresses
        }
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)

	def __repr__(self):
		return str(self)
		
	def __str__(self):
		t  = '== INTERFACE ==\r\n'
		t += 'Name: %s\r\n' % self.ifname
		t += 'ifindex: %s\r\n' % self.ifindex
		for addr in self.addresses:
			t += 'Address: %s\r\n' % str(addr)
		return t




import asyncio


async def generic_read(reader, n):
	return reader.read(n)


async def generic_write(writer, data):
	writer.write(data)
	await writer.drain()


async def readexactly_or_exc(reader, n, timeout = None):
	"""
	Helper function to read exactly N amount of data from the wire.
	:param reader: The reader object
	:type reader: asyncio.StreamReader
	:param n: The maximum amount of bytes to read.
	:type n: int
	:param timeout: Time in seconds to wait for the reader to return data
	:type timeout: int
	:return: bytearray
	"""
	temp = await asyncio.gather(*[asyncio.wait_for(reader.readexactly(n), timeout = timeout)], return_exceptions=True)
	if isinstance(temp[0], bytes):
		return temp[0]
	else:
		raise temp[0]

	return data


async def read_or_exc(reader, n, timeout = None):
	"""
	Helper function to read N amount of data from the wire.
	:param reader: The reader object
	:type reader: asyncio.StreamReader
	:param n: The maximum amount of bytes to read. BEWARE: this only sets an upper limit of the data to be read
	:type n: int
	:param timeout: Time in seconds to wait for the reader to return data
	:type timeout: int
	:return: bytearray
	"""

	temp = await asyncio.gather(*[asyncio.wait_for(reader.read(n), timeout = timeout)], return_exceptions=True)
	if isinstance(temp[0], bytes):
		return temp[0]
	else:
		raise temp[0]

	return data


async def readuntil_or_exc(reader, pattern, timeout = None):
	"""
	Helper function to read the wire until a certain pattern is reached.
	:param reader: The reader object
	:type reader: asyncio.StreamReader
	:param pattern: The pattern marking the end of read
	:type pattern: bytearray
	:param timeout: Time in seconds to wait for the reader to reach the pattern
	:type timeout: int
	:return: bytearray
	"""

	temp = await asyncio.gather(*[asyncio.wait_for(reader.readuntil(pattern), timeout = timeout)], return_exceptions=True)
	if isinstance(temp[0], bytes):
		return temp[0]
	else:
		raise temp[0]



async def readline_or_exc(reader, timeout = None):
	"""
	Helper function to read the wire until an end-of-line character is reached.
	:param reader: The reader object
	:type reader: asyncio.StreamReader
	:param timeout: Time in seconds to wait for the reader to reach the pattern
	:type timeout: int
	:return: bytearray
	"""

	temp = await asyncio.gather(*[asyncio.wait_for(reader.readline(), timeout = timeout)], return_exceptions=True)
	if isinstance(temp[0], bytes):
		return temp[0]
	else:
		raise temp[0]


async def sendall(writer, data):
	"""
	Helper function that writes all the data to the wire
	:param writer: Writer object
	:type writer: asyncio.StreamWriter
	:param data: Data to be written
	:type data: bytearray
	:return: None
	"""
	try:
		writer.write(data)
		await writer.drain()
	except Exception as e:
		raise R3ConnectionClosed()


class R3ConnectionClosed(Exception):
	pass


@asyncio.coroutine
async def wait_mp_event(event, aio_event):
	event.wait(timeout = None)
	aio_event.set()
	return



import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.TNS import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession

class TNSSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.parser = TNSPacket
		self.firstpacket = True

	def __repr__(self):
		t = '== TNS Session ==\r\n'
		return t


class TNS(ResponderServer):
	def init(self):
		pass

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):

		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				msg = result[0]

			print(msg)
			if self.session.firstpacket == True:
				ta = TNSResend()
				packet = TNSPacket()
				packet.payload = ta

				await self.send_data(packet.to_bytes())
				self.session.firstpacket = False
				continue

			if msg.header.packet_type == TNSPacketType.CONNECT:
				ta = TNSAccept()
				ta.version = 0x0134
				ta.service_flags = 0x0000
				ta.sdu_size = 0x0800
				ta.maximum_tdu_size = 0x7fff
				ta.byte_order = 256
				ta.data_length = None
				ta.data_offset = 24
				ta.flags1 = 65
				ta.flags2 = 1
				ta.padding = None
				ta.data = None
				packet = TNSPacket()
				packet.payload = ta

				await self.send_data(packet.to_bytes())

			elif msg.header.packet_type == TNSPacketType.DATA:
				t = "deadbeef00920a2001000004000004000300000000000400050a200100000800" +\
					"0100000b58884d7db000120001deadbeef000300000004000400010001000200" +\
					"01000300000000000400050a20010000020003e0e100020006fcff0002000200" +\
					"000000000400050a200100000c0001001106100c0f0a0b080201030003000200" +\
					"000000000400050a20010000030001000301"

				ta = TNSData()
				ta.flags = 0
				ta.data = bytes.fromhex(t)
				packet = TNSPacket()
				packet.payload = ta

				await self.send_data(packet.to_bytes())


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

# https://tools.ietf.org/html/rfc4511
# https://msdn.microsoft.com/en-us/library/cc223501.aspx
# https://ldap3.readthedocs.io/bind.html

from asn1crypto import core
import enum
import os

from responder3.core.commons import *
from responder3.core.logging.log_objects import Credential
from responder3.core.asyncio_helpers import *

TAG = 'explicit'

# class
UNIVERSAL = 0
APPLICATION = 1
CONTEXT = 2

#https://msdn.microsoft.com/en-us/library/cc223359.aspx
class MSLDAPCapabilities(core.ObjectIdentifier):
    _map = {
		'1.2.840.113556.1.4.800' : 'LDAP_CAP_ACTIVE_DIRECTORY_OID',
		'1.2.840.113556.1.4.1791': 'LDAP_CAP_ACTIVE_DIRECTORY_LDAP_INTEG_OID',
		'1.2.840.113556.1.4.1670': 'LDAP_CAP_ACTIVE_DIRECTORY_V51_OID',
		'1.2.840.113556.1.4.1880': 'LDAP_CAP_ACTIVE_DIRECTORY_ADAM_DIGEST_OID',
		'1.2.840.113556.1.4.1851': 'LDAP_CAP_ACTIVE_DIRECTORY_ADAM_OID',
		'1.2.840.113556.1.4.1920': 'LDAP_CAP_ACTIVE_DIRECTORY_PARTIAL_SECRETS_OID',
		'1.2.840.113556.1.4.1935': 'LDAP_CAP_ACTIVE_DIRECTORY_V60_OID',
		'1.2.840.113556.1.4.2080': 'LDAP_CAP_ACTIVE_DIRECTORY_V61_R2_OID',
		'1.2.840.113556.1.4.2237': 'LDAP_CAP_ACTIVE_DIRECTORY_W8_OID',
	}

class scope(core.Enumerated):
	_map = {
		0 : 'baseObject',
		1 : 'singleLevel',
		2 : 'wholeSubtree',
	}

class derefAliases(core.Enumerated):
	_map = {
		0 : 'neverDerefAliases',
		1 : 'derefInSearching',
		2 : 'derefFindingBaseObj',
		3 : 'derefAlways',
	}

class resultCode(core.Enumerated):
	_map = {
		0   : 'success',
        1  : 'operationsError',
        2  : 'protocolError',
        3  : 'timeLimitExceeded',
        4  : 'sizeLimitExceeded',
        5  : 'compareFalse',
        6  : 'compareTrue',
        7  : 'authMethodNotSupported',
        8  : 'strongerAuthRequired',
        10 : 'referral',
        11 : 'adminLimitExceeded',
        12 : 'unavailableCriticalExtension',
        13 : 'confidentialityRequired',
        14 : 'saslBindInProgress',
        16 : 'noSuchAttribute',
        17 : 'undefinedAttributeType',
        18 : 'inappropriateMatching',
        19 : 'constraintViolation',
        20 : 'attributeOrValueExists',
        21 : 'invalidAttributeSyntax',
        32 : 'noSuchObject',
        33 : 'aliasProblem',
        34 : 'invalidDNSyntax',
        36 : 'aliasDereferencingProblem',
        48 : 'inappropriateAuthentication',
        49 : 'invalidCredentials',
        50 : 'insufficientAccessRights',
        51 : 'busy',
        52 : 'unavailable',
        53 : 'unwillingToPerform',
        54 : 'loopDetect',
        64 : 'namingViolation',
        65 : 'objectClassViolation',
        66 : 'notAllowedOnNonLeaf',
        67 : 'notAllowedOnRDN',
        68 : 'entryAlreadyExists',
        69 : 'objectClassModsProhibited',
        71 : 'affectsMultipleDSAs',
        80 : 'other',
	}
	
	
class LDAPString(core.OctetString):
	pass
	
class LDAPDN(core.OctetString):
	pass
	
class LDAPOID(core.OctetString):
	pass
	
class URI(LDAPString):
	pass
	
class Referral(core.SequenceOf):
	_child_spec = URI

class Control(core.Sequence):
	_fields = [
		('controlType', LDAPOID, {'tag_type': TAG, 'tag': 0}),
		('criticality', core.Boolean, {'tag_type': TAG, 'tag': 1}),
		('controlValue', core.OctetString, {'tag_type': TAG, 'tag': 2, 'optional': True}),	
	]
	
class Controls(core.SequenceOf):
	_child_spec = Control
	
class SaslCredentials(core.Sequence):
	_fields = [
		('mechanism', LDAPString),
		('credentials', core.OctetString, {'optional': True}),	
	]
	
class SicilyPackageDiscovery(core.OctetString):
	pass

class SicilyNegotiate(core.OctetString):
	pass

class SicilyResponse(core.OctetString):
	pass
	
class AuthenticationChoice(core.Choice):
	_alternatives = [
		('simple', core.OctetString, {'implicit': (CONTEXT, 0)}),
		('sasl', SaslCredentials, {'implicit': (CONTEXT, 3)}),
		('sicily_disco', SicilyPackageDiscovery, {'implicit': (CONTEXT, 9)}), # FUCK
		('sicily_nego', SicilyNegotiate, {'implicit': (CONTEXT, 10)}), #YOU
		('sicily_resp', SicilyResponse, {'implicit': (CONTEXT, 11)}), #MICROSOFT
	]

class BindRequest(core.Sequence):
	_fields = [
		('version', core.Integer),
		('name', core.OctetString),
		('authentication', AuthenticationChoice),	
	]
	
class BindResponse(core.Sequence):
	_fields = [
		('resultCode', resultCode),
		('matchedDN', LDAPDN),
		('diagnosticMessage', LDAPString),
		('referral', Referral, {'optional': True}),
		('serverSaslCreds', core.OctetString, {'optional': True}),
	]

class AttributeDescription(LDAPString):
	pass

class AttributeValue(core.OctetString):
	pass


class MatchingRuleId(LDAPString):
	pass

class AssertionValue(core.OctetString):
	pass

class AttributeValueAssertion(core.Sequence):
	_fields = [
		('attributeDesc', AttributeDescription),
		('assertionValue', AssertionValue),
	]

class SubString(core.Choice):
	_alternatives = [
		('initial', AssertionValue, {'implicit': (APPLICATION , 0) }  ),
		('any', AssertionValue, {'implicit': (APPLICATION , 1) }  ),
		('final', AssertionValue, {'implicit': (APPLICATION , 2) }  ),
	]


class SubstringFilter(core.Sequence):
	_fields = [
		('type', AttributeDescription),
		('substrings', AssertionValue),
	]

class MatchingRuleAssertion(core.Sequence):
	_fields = [
		('matchingRule', MatchingRuleId, {'implicit': (APPLICATION, 1), 'optional' : True}  ),
		('type', AttributeDescription, {'implicit': (APPLICATION, 2), 'optional' : True}  ),
		('matchValue', AssertionValue, {'implicit': (APPLICATION, 3), 'optional' : False}  ),
		('dnAttributes', core.Boolean, {'implicit': (APPLICATION, 4), 'optional' : False}  ),
	]



class Filter(core.Choice):
	_alternatives = [
		#('and', Filters, {'implicit': (CONTEXT , 0) }  ),
		#('or', Filters, {'implicit': (CONTEXT , 1) }  ),
		#('not', Filter, {'implicit': (CONTEXT , 2) }  ),
		('equalityMatch', AttributeValueAssertion, {'implicit': (CONTEXT , 3) }  ),
		('substrings', SubstringFilter, {'implicit': (CONTEXT , 4) }  ),
		('greaterOrEqual', AttributeValueAssertion, {'implicit': (CONTEXT , 5) }  ),
		('lessOrEqual', AttributeValueAssertion, {'implicit': (CONTEXT , 6) }  ),
		('present', AttributeDescription, {'implicit': (CONTEXT , 7) }  ),
		('approxMatch', AttributeValueAssertion, {'implicit': (CONTEXT , 8) }  ),
		('extensibleMatch', MatchingRuleAssertion, {'implicit': (CONTEXT , 9) }  ),

	]
class Filters(core.SequenceOf):
	_child_spec = Filter

class AttributeSelection(core.SequenceOf):
	_child_spec = LDAPString

class SearchRequest(core.Sequence):
	_fields = [
		('baseObject', LDAPDN),
		('scope', scope),
		('derefAliases', derefAliases),
		('sizeLimit', core.Integer),
		('timeLimit', core.Integer),
		('typesOnly', core.Boolean),
		('filter', Filter),
		('attributes', AttributeSelection),
	]

class AttributeValueSet(core.SetOf):
	_child_spec = AttributeValue


class PartialAttribute(core.Sequence):
	_fields = [
		('type', AttributeDescription),
		('attributes', AttributeValueSet),
	]

class PartialAttributeList(core.SequenceOf):
	_child_spec = PartialAttribute

class SearchResultEntry(core.Sequence):
	_fields = [
		('objectName', LDAPDN),
		('attributes', PartialAttributeList),
	]

class SearchResultReference(core.SequenceOf):
	_child_spec = URI

class UnbindRequest(core.Null):
	pass

class LDAPResult(core.Sequence):
	_fields = [
		('resultCode', resultCode ),
		('matchedDN', LDAPDN),
		('diagnosticMessage', LDAPString),
		('referral', Referral,  {'implicit': (CONTEXT, 3), 'optional': True}),
	]

class SearchResultDone(LDAPResult):
	pass

class protocolOp(core.Choice):
	_alternatives = [
		('bindRequest', BindRequest, {'implicit': (APPLICATION , 0) }  ),
		('bindResponse', BindResponse, {'implicit': (APPLICATION , 1) }  ),
		('unbindRequest', UnbindRequest, {'implicit': (APPLICATION,2) }  ),
		('searchRequest', SearchRequest, {'implicit': (APPLICATION,3) }  ),
		('searchResEntry', SearchResultEntry, {'implicit': (APPLICATION,4) }  ),
		('searchResDone', SearchResultDone, {'implicit': (APPLICATION,5) }  ),
		('searchResRef', SearchResultReference, {'implicit': (APPLICATION,6) }  ),
		#('modifyRequest', ModifyRequest, {'implicit': (APPLICATION,0) }  ),
		#('modifyResponse', ModifyResponse, {'implicit': (APPLICATION,0) }  ),
		#('addRequest', AddRequest, {'implicit': (APPLICATION,0) }  ),
		#('addResponse', AddResponse, {'implicit': (APPLICATION,0) }  ),
		#('delRequest', DelRequest, {'implicit': (APPLICATION,0) }  ),
		#('delResponse', DelResponse, {'implicit': (APPLICATION,0) }  ),
		#('modDNRequest', ModifyDNRequest, {'implicit': (APPLICATION,0) }  ),
		#('modDNResponse', ModifyDNResponse, {'implicit': (APPLICATION,0) }  ),
		#('compareRequest', CompareRequest, {'implicit': (APPLICATION,0) }  ),
		#('compareResponse', CompareResponse, {'implicit': (APPLICATION,0) }  ),
		#('abandonRequest', AbandonRequest, {'implicit': (APPLICATION,0) }  ),
		#('extendedReq', ExtendedRequest, {'implicit': (APPLICATION,0) }  ),
		#('abandonRequest', AbandonRequest, {'implicit': (APPLICATION,0) }  ),
		#('extendedResp', ExtendedResponse, {'implicit': (APPLICATION,0) }  ),
	]
	
class LDAPMessage(core.Sequence):
	_fields = [
		('messageID', core.Integer),
		('protocolOp', protocolOp),
		('controls', Controls, {'implicit': (CONTEXT, 0), 'optional': True}),	
	]
	




class LdapParser:
	def __init__(self):
		pass
	
	@staticmethod
	async def from_streamreader(reader):
		preread = 6
		lb = await readexactly_or_exc(reader, preread)
		
		remaining_length = LdapParser.calcualte_length(lb) - preread
		remaining_data = await readexactly_or_exc(reader, remaining_length)
		
		message = LDAPMessage.load(lb+remaining_data)
		
		return message
		
	def calcualte_length(data):
		"""
		LDAP protocol doesnt send the total length of the message in the header,
		it only sends raw ASN1 encoded data structures, which has the length encoded.
		This function "decodes" the length os the asn1 structure, and returns it as int.
		"""
		if data[1] <= 127:
			return data[1] + 2
		else:
			bcount = data[1] - 128
			if (bcount +2 ) > len(data):
				raise Exception('LDAP data too larage! Length byte count: %s' % bcount)
			return int.from_bytes(data[2:2+bcount], byteorder = 'big', signed = False) + bcount + 2
			




import logging
import traceback
import sys
import io
import asyncio

from responder3.core.logging.log_objects import *
from responder3.core.commons import Connection, ConnectionStatus

class Logger:
	"""
	This class is used to provie a better logging experience for asyncio based classes/functions
	Probably will replace logtask "solution" with this one in the future
	TODO
	"""
	def __init__(self, name, logger = None, logQ = None, level = logging.DEBUG, connection = None):
		self.level = level
		self.consumers = {}
		self.logger = logger
		self.name = name
		self._connection = connection

		self.is_final = True
		if logQ:
			self.logQ = logQ
			self.is_final = False
		else:
			self.logQ = asyncio.Queue()
		
	async def run(self):
		"""
		you only need to call this function IF the logger instance is the final dst!
		Also, consumers will not work if this is not the final one!
		"""
		if self.is_final == False:
			return
		try:
			while True:
				logmsg = await self.logQ.get()
				await self.handle_logger(logmsg)
				if len(self.consumers) > 0:
					await self.handle_consumers(logmsg)
		
		except Exception as e:
			print('Logger run exception! %s' % e)
			
	async def handle_logger(self, msg):
		print('%s %s %s %s' % (datetime.datetime.utcnow().isoformat(), self.name, level, msg))
		
	async def handle_consumers(self, msg):
		try:
			for consumer in self.consumers:
				await consumer.process_log(msg)
		except Exception as e:
			print(e)
			
	async def debug(self, msg):
		await self.logQ.put(LogEntry(logging.DEBUG, self.name, msg, self._connection))
		
	async def info(self, msg):
		await self.logQ.put(LogEntry(logging.INFO, self.name, msg, self._connection))
	
	async def exception(self, message = None):
		sio = io.StringIO()
		ei = sys.exc_info()
		tb = ei[2]
		traceback.print_exception(ei[0], ei[1], tb, None, sio)
		msg = sio.getvalue()
		if msg[-1] == '\n':
			msg = msg[:-1]
		sio.close()
		if message is not None:
			msg = '%s : %s' % (message,msg)
		await self.logQ.put(LogEntry(logging.ERROR, self.name, msg, self._connection))
			
	async def error(self, msg):
		await self.logQ.put(LogEntry(logging.ERROR, self.name, msg, self._connection))
		
	async def warning(self, msg):
		await self.logQ.put(LogEntry(logging.WARNING, self.name, msg, self._connection))
		
	async def log(self, level, msg):
		"""
		Level MUST be bigger than 0!!!
		"""
		await self.logQ.put(LogEntry(level, self.name, msg, self._connection))

	async def connection(self, connection, status):
		"""
		Logs incoming connection
		:param connection: The Connection object to log
		:type connection: Connection
		:param status: Connection status
		:type: ConnectionStatus
		:return: None
		"""
		if status == ConnectionStatus.OPENED or status == ConnectionStatus.STATELESS:
			#await self.info('New connection opened from %s:%d' % (connection.remote_ip, connection.remote_port))
			co = ConnectionOpened(connection, self.name)
			await self.logQ.put(co)
			
		elif status == ConnectionStatus.CLOSED:
			#await self.info('Connection closed by %s:%d' % (connection.remote_ip, connection.remote_port))
			cc = ConnectionClosed(connection, self.name)
			await self.logQ.put(cc)

	async def credential(self, credential):
		"""
		Create a credential message and send it to the LogProcessor for procsesing
		"""
		credential.module = self.name
		credential.connection = self._connection
		credential.client_addr = self._connection.remote_ip
		credential.client_rdns = self._connection.remote_dns
		await self.logQ.put(credential)

	
	async def poisonresult(self, mode, requestName = None, poisonName = None, poisonIP = None, request_type = None):
		pr = PoisonResult(self._connection)
		pr.module = self.name
		pr.target = self._connection.remote_ip
		pr.request_name = requestName
		pr.request_type = request_type
		pr.poison_name = poisonName
		pr.poison_addr = poisonIP
		pr.mode = mode

		await self.logQ.put(pr)
	
	"""
	async def log_email(self, emailEntry):
		await self.logger.email(emailEntry)
	"""

	async def traffic(self, traffic):
		traffic.module = self.name
		traffic.connection = self._connection
		await self.logQ.put(traffic)

	async def proxy(self, data, laddr, raddr, is_modified = False):
		msg = '[O]' if is_modified == False else '[M]'
		msg += '[%s:%s -> %s:%s] %s' % (str(laddr[0]), str(laddr[1]) , str(raddr[0]), str(raddr[1]), data.hex())
		await self.logQ.put(LogEntry(logging.INFO, self.name, msg, self._connection))

	async def proxydata(self, data, laddr, raddr, is_ssl, data_type, proto = 'TCP'):
		proxydata = ProxyData()
		proxydata.src_addr  = laddr
		proxydata.dst_addr  = raddr
		proxydata.proto     = proto
		proxydata.isSSL     = is_ssl
		proxydata.timestamp = datetime.datetime.utcnow()
		proxydata.data_type = data_type
		proxydata.data      = data
		proxydata.module = self.name
		proxydata.connection = self._connection
		await self.logQ.put(proxydata)

		
	def add_consumer(self, consumer):
		self.consumers[consumer] = 0
		
	def del_consumer(self, consumer):
		if consumer in self.consumers:
			del self.consumers[consumer]
			
			
def r3exception(funct):
	"""
	Decorator for handling exceptions
	Use it with the Logger class only!!!
	"""
	async def wrapper(*args, **kwargs):
		this = args[0] #renaming self to 'this'
		try:
			t = await funct(*args, **kwargs)
			return t
		except Exception as e:
			await this.logger.exception(funct.__name__)
			return
			
	return wrapper

def r3trafficlogexception(funct):
	"""
	Decorator for handling exceptions
	Use it with the Logger class only!!!
	"""
	async def wrapper(*args, **kwargs):
		this = args[0] #renaming self to 'this'
		try:
			t = await funct(*args, **kwargs)
			return t
		except asyncio.CancelledError:
			await this.logger.debug('Got cancelled! Probably timeout')
		except:				
			await this.logger.exception(funct.__name__)
			traffic_in  = await this.creader.log_comms()
			traffic_out = await this.cwriter.log_comms()
			traffic_in.data_sent = traffic_out.data_sent
			await this.logger.traffic(traffic_in)

			if hasattr(this, 'shutdown_evt') == True:
				this.shutdown_evt.set()
			
			return
			
	return wrapper


import re
import socket
import struct
import logging
import asyncio
import ipaddress
import traceback

from responder3.core.sockets import setup_base_socket
from responder3.core.commons import PoisonerMode, ResponderPlatform
from responder3.protocols.LLMNR import * 
from responder3.protocols.DNS import * 
from responder3.core.servertemplate import ResponderServer, ResponderServerSession, ResponderServerGlobalSession


class LLMNRGlobalSession(ResponderServerGlobalSession):
	def __init__(self, listener_socket_config, settings, log_queue):
		ResponderServerGlobalSession.__init__(self, log_queue, self.__class__.__name__)
		self.listener_socket_config = listener_socket_config
		self.settings = settings

		self.spooftable = []
		self.poisonermode = PoisonerMode.ANALYSE

		self.parse_settings()

	def parse_settings(self):
		if self.settings is None:
			self.logger.info('No settings defined, adjusting to Analysis functionality!')
		else:
			# parse the poisoner mode
			if isinstance(self.settings['mode'], str):
				self.poisonermode = PoisonerMode[self.settings['mode'].upper()]

			# compiling re strings to actual re objects and converting IP strings to IP objects
			if self.poisonermode == PoisonerMode.SPOOF:
				for exp in self.settings['spooftable']:
					self.spooftable.append((re.compile(exp),ipaddress.ip_address(self.settings['spooftable'][exp])))


class LLMNRSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)


class LLMNR(ResponderServer):
	@staticmethod
	def custom_socket(socket_config):
		if socket_config.bind_family == 4:
			ip = ipaddress.ip_address('224.0.0.252')
			sock = setup_base_socket(
				socket_config,
				bind_ip_override = ipaddress.ip_address('0.0.0.0') #if socket_config.platform != ResponderPlatform.WINDOWS else None
			)
			sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 255)
			mreq = struct.pack("=4sl", ip.packed, socket.INADDR_ANY)
			sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)			

		else:
			ip = ipaddress.ip_address('FF02:0:0:0:0:0:1:3')
			sock = setup_base_socket(
				socket_config,
				bind_ip_override = ipaddress.ip_address('::') if socket_config.platform != ResponderPlatform.WINDOWS else None
			)
			sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 255)
			sock.setsockopt(
				41 if socket_config.platform == ResponderPlatform.WINDOWS else socket.IPPROTO_IPV6,
				socket.IPV6_JOIN_GROUP,
				struct.pack('16sI', ip.packed, socket_config.bind_iface_idx)
			)

		return sock

	def init(self):
		self.parser = LLMNRPacket

	async def parse_message(self):
		msg = await asyncio.wait_for(self.parser.from_streamreader(self.creader), timeout=1)
		return msg

	async def run(self):
		try:
			msg = await asyncio.wait_for(self.parse_message(), timeout=1)
			if self.globalsession.poisonermode == PoisonerMode.ANALYSE:
				for q in msg.Questions:
					await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, request_type = q.QTYPE.name)

			else:
				answers = []
				for targetRE, ip in self.globalsession.spooftable:
					for q in msg.Questions:
						if targetRE.match(q.QNAME.name):
							await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, poisonName = str(targetRE), poisonIP = ip, request_type = q.QTYPE.name)
							res = None
							if ip.version == 4 and q.QTYPE == DNSType.A:
								res = DNSAResource.construct(q.QNAME.name, ip)
							elif ip.version == 6 and q.QTYPE == DNSType.AAAA:
								res = DNSAAAAResource.construct(q.QNAME.name, ip)

							if res:
								answers.append(res)
						else:
							await self.logger.poisonresult(PoisonerMode.ANALYSE, requestName = q.QNAME.name, request_type = q.QTYPE.name)
				
				response = LLMNRPacket.construct(TID = msg.TransactionID,
									 response = LLMNRResponse.RESPONSE, 
									 answers = answers,
									 questions = msg.Questions
								  )
				self.cwriter.write(response.to_bytes())

		except Exception as e:
			await self.logger.exception()


